@@ -27,6 +27,15 @@
             "On the difficulty of evaluating baselines:
             A study on recommender systems."
             arXiv preprint arXiv:1905.01395 (2019).
+        (f) Franc, Vojtech, Vaclav Hlavac, and Mirko Navara.
+            "Sequential coordinate-wise algorithm for the
+            non-negative least squares problem."
+            International Conference on Computer Analysis of Images
+            and Patterns. Springer, Berlin, Heidelberg, 2005.
+        (e) Zhou, Yunhong, et al.
+            "Large-scale parallel collaborative filtering for the netflix prize."
+            International conference on algorithmic applications in management.
+            Springer, Berlin, Heidelberg, 2008.
 
     For information about the models offered here and how they are fit to
     the data, see the files 'collective.c' and 'offsets.c'.
@@ -40,7 +49,7 @@
 
     MIT License:
 
-    Copyright (c) 2020 David Cortes
+    Copyright (c) 2021 David Cortes
 
     All rights reserved.
 
@@ -73,7 +82,6 @@ extern "C" {
 #include <string.h>
 #include <stdbool.h>
 #include <math.h>
-#include <signal.h>
 #ifndef _FOR_R
     #include <stdio.h>
 #endif
@@ -82,6 +90,11 @@ extern "C" {
 #else
     #define omp_get_thread_num() (0)
 #endif
+#include <signal.h>
+typedef void (*sig_t_)(int);
+#if !defined(_WIN32) && !defined(_WIN64) && !defined(_MSC_VER)
+    #include <unistd.h>
+#endif
 
 #ifdef _FOR_PYTHON
     /* This contains the standard cblas.h header */
@@ -147,6 +160,8 @@ extern "C" {
     #define real_t double
     #define exp_t exp
     #define log_t log
+    #define fabs_t fabs
+    #define InfVal HUGE_VAL
     #define cblas_tdot cblas_ddot
     #define cblas_tcopy cblas_dcopy
     #define cblas_taxpy cblas_daxpy
@@ -168,6 +183,8 @@ extern "C" {
     #define real_t float
     #define exp_t expf
     #define log_t logf
+    #define fabs_t fabsf
+    #define InfVal HUGE_VALF
     #define cblas_tdot cblas_sdot
     #define cblas_tcopy cblas_scopy
     #define cblas_taxpy cblas_saxpy
@@ -186,6 +203,10 @@ extern "C" {
     #define tgelsd_ sgelsd_
 #endif
 
+#ifndef isfinite
+    #define isfinite(x) ((x) > -InfVal && (x) < InfVal)
+#endif
+
 #if !defined(USE_INT64) && !defined(MKL_ILP64)
     #define int_t int
 #else
@@ -405,10 +426,14 @@ void append_ones_last_col
     real_t *restrict outp
 );
 void fill_lower_triangle(real_t A[], size_t n, size_t lda);
+void print_err_msg(const char *msg);
 void print_oom_message(void);
+void act_on_interrupt(int retval, bool handle_interrupt);
 #ifdef _FOR_R
 void R_nan_to_C_nan(real_t arr[], size_t n);
 #endif
+long double compensated_sum(real_t *arr, size_t n);
+long double compensated_sum_product(real_t *restrict arr1, real_t *restrict arr2, size_t n);
 
 
 /* common.c */
@@ -439,11 +464,15 @@ void factors_closed_form
     real_t *restrict weight,
     real_t *restrict buffer_real_t,
     real_t lam, real_t lam_last,
+    real_t l1_lam, real_t l1_lam_last,
+    bool scale_lam,
     real_t *restrict precomputedTransBtBinvBt,
     real_t *restrict precomputedBtB, int_t cnt_NA, int_t ld_BtB,
     bool BtB_has_diag, bool BtB_is_scaled, real_t scale_BtB, int_t n_BtB,
     real_t *restrict precomputedBtBchol, bool NA_as_zero,
-    bool use_cg, int_t max_cg_steps, /* <- 'cg' should not be used for new data */
+    bool use_cg, int_t max_cg_steps,/* <- 'cg' should not be used for new data*/
+    bool nonneg, int_t max_cd_steps,
+    real_t *restrict bias_BtX, real_t *restrict bias_X, real_t bias_X_glob,
     bool force_add_diag
 );
 void factors_explicit_cg
@@ -463,6 +492,7 @@ void factors_explicit_cg_NA_as_zero_weighted
     real_t *restrict Xa, int_t ixB[], size_t nnz,
     real_t *restrict weight,
     real_t *restrict precomputedBtB, int_t ld_BtB,
+    real_t *restrict bias_BtX, real_t *restrict bias_X, real_t bias_X_glob,
     real_t *restrict buffer_real_t,
     real_t lam, real_t lam_last,
     int_t max_cg_steps
@@ -493,10 +523,51 @@ void factors_implicit_chol
     real_t *restrict a_vec, int_t k,
     real_t *restrict B, size_t ldb,
     real_t *restrict Xa, int_t ixB[], size_t nnz,
-    real_t lam,
+    real_t lam, real_t l1_lam,
     real_t *restrict precomputedBtB, int_t ld_BtB,
+    bool nonneg, int_t max_cd_steps,
     real_t *restrict buffer_real_t
 );
+void solve_nonneg
+(
+    real_t *restrict BtB,
+    real_t *restrict BtX, /* <- solution will be here */
+    real_t *restrict buffer_real_t,
+    int_t k,
+    real_t l1_lam, real_t l1_lam_last,
+    size_t max_cd_steps,
+    bool fill_lower
+);
+void solve_nonneg_batch
+(
+    real_t *restrict BtB,
+    real_t *restrict BtX, /* <- solution will be here */
+    real_t *restrict buffer_real_t,
+    int_t m, int_t k, size_t lda,
+    real_t l1_lam, real_t l1_lam_last,
+    size_t max_cd_steps,
+    int nthreads
+);
+void solve_elasticnet
+(
+    real_t *restrict BtB,
+    real_t *restrict BtX, /* <- solution will be here */
+    real_t *restrict buffer_real_t,
+    int_t k,
+    real_t l1_lam, real_t l1_lam_last,
+    size_t max_cd_steps,
+    bool fill_lower
+);
+void solve_elasticnet_batch
+(
+    real_t *restrict BtB,
+    real_t *restrict BtX, /* <- solution will be here */
+    real_t *restrict buffer_real_t,
+    int_t m, int_t k, size_t lda,
+    real_t l1_lam, real_t l1_lam_last,
+    size_t max_cd_steps,
+    int nthreads
+);
 real_t fun_grad_Adense
 (
     real_t *restrict g_A,
@@ -555,6 +626,7 @@ size_t buffer_size_optimizeA
 (
     size_t n, bool full_dense, bool near_dense, bool do_B,
     bool has_dense, bool has_weights, bool NA_as_zero,
+    bool nonneg, bool has_l1,
     size_t k, size_t nthreads,
     bool pass_allocated_BtB, bool keep_precomputedBtB,
     bool use_cg, bool finalize_chol
@@ -563,6 +635,7 @@ size_t buffer_size_optimizeA_implicit
 (
     size_t k, size_t nthreads,
     bool pass_allocated_BtB,
+    bool nonneg, bool has_l1,
     bool use_cg, bool finalize_chol
 );
 void optimizeA
@@ -574,9 +647,13 @@ void optimizeA
     real_t *restrict Xfull, int_t ldX, bool full_dense, bool near_dense,
     int_t cnt_NA[], real_t *restrict weight, bool NA_as_zero,
     real_t lam, real_t lam_last,
+    real_t l1_lam, real_t l1_lam_last,
+    bool scale_lam,
     bool do_B, bool is_first_iter,
     int_t nthreads,
     bool use_cg, int_t max_cg_steps,
+    bool nonneg, int_t max_cd_steps,
+    real_t *restrict bias_BtX, real_t *restrict bias_X, real_t bias_X_glob,
     bool keep_precomputedBtB,
     real_t *restrict precomputedBtB, bool *filled_BtB,
     real_t *restrict buffer_real_t
@@ -587,23 +664,26 @@ void optimizeA_implicit
     real_t *restrict B, size_t ldb,
     int_t m, int_t n, int_t k,
     size_t Xcsr_p[], int_t Xcsr_i[], real_t *restrict Xcsr,
-    real_t lam,
+    real_t lam, real_t l1_lam,
     int_t nthreads,
     bool use_cg, int_t max_cg_steps, bool force_set_to_zero,
-    real_t *restrict precomputedBtB,
+    bool nonneg, int_t max_cd_steps,
+    real_t *restrict precomputedBtB, /* <- will be calculated if not passed */
     real_t *restrict buffer_real_t
 );
 int_t initialize_biases
 (
     real_t *restrict glob_mean, real_t *restrict biasA, real_t *restrict biasB,
-    bool user_bias, bool item_bias,
+    bool user_bias, bool item_bias, bool center,
     real_t lam_user, real_t lam_item,
+    bool scale_lam,
     int_t m, int_t n,
     int_t m_bias, int_t n_bias,
     int_t ixA[], int_t ixB[], real_t *restrict X, size_t nnz,
     real_t *restrict Xfull, real_t *restrict Xtrans,
     size_t Xcsr_p[], int_t Xcsr_i[], real_t *restrict Xcsr,
     size_t Xcsc_p[], int_t Xcsc_i[], real_t *restrict Xcsc,
+    bool nonneg,
     int_t nthreads
 );
 int_t center_by_cols
@@ -657,12 +737,30 @@ int_t fit_most_popular
     real_t *restrict biasA, real_t *restrict biasB,
     real_t *restrict glob_mean,
     real_t lam_user, real_t lam_item,
+    bool scale_lam,
     real_t alpha,
     int_t m, int_t n,
     int_t ixA[], int_t ixB[], real_t *restrict X, size_t nnz,
     real_t *restrict Xfull,
     real_t *restrict weight,
     bool implicit, bool adjust_weight, bool apply_log_transf,
+    bool nonneg,
+    real_t *restrict w_main_multiplier,
+    int_t nthreads
+);
+int_t fit_most_popular_internal
+(
+    real_t *restrict biasA, real_t *restrict biasB,
+    real_t *restrict glob_mean, bool center,
+    real_t lam_user, real_t lam_item,
+    bool scale_lam,
+    real_t alpha,
+    int_t m, int_t n,
+    int_t ixA[], int_t ixB[], real_t *restrict X, size_t nnz,
+    real_t *restrict Xfull,
+    real_t *restrict weight,
+    bool implicit, bool adjust_weight, bool apply_log_transf,
+    bool nonneg,
     real_t *restrict w_main_multiplier,
     int_t nthreads
 );
@@ -802,12 +900,16 @@ void collective_closed_form_block
     real_t *restrict Xones, int_t incXones,
     real_t *restrict weight,
     real_t lam, real_t w_user, real_t w_implicit, real_t lam_last,
+    real_t l1_lam, real_t l1_lam_bias,
+    bool scale_lam, bool scale_lam_sideinfo,
     real_t *restrict precomputedBtB, int_t cnt_NA_x,
     real_t *restrict precomputedCtCw, int_t cnt_NA_u,
     real_t *restrict precomputedBeTBeChol, int_t n_BtB,
     real_t *restrict precomputedBiTBi,
     bool add_X, bool add_U,
     bool use_cg, int_t max_cg_steps,/* <- 'cg' should not be used for new data*/
+    bool nonneg, int_t max_cd_steps,
+    real_t *restrict bias_BtX, real_t *restrict bias_X, real_t bias_X_glob,
     real_t *restrict buffer_real_t
 );
 void collective_closed_form_block_implicit
@@ -819,13 +921,14 @@ void collective_closed_form_block_implicit
     real_t *restrict u_vec, int_t cnt_NA_u,
     real_t *restrict u_vec_sp, int_t u_vec_ixB[], size_t nnz_u_vec,
     bool NA_as_zero_U,
-    real_t lam, real_t w_user,
+    real_t lam, real_t l1_lam, real_t w_user,
     real_t *restrict precomputedBeTBe,
     real_t *restrict precomputedBtB, /* for cg, should NOT have lambda added */
     real_t *restrict precomputedBeTBeChol,
     real_t *restrict precomputedCtCw,
     bool add_U, bool shapes_match,
-    bool use_cg, int_t max_cg_steps,
+    bool use_cg, int_t max_cg_steps,/* <- 'cg' should not be used for new data*/
+    bool nonneg, int_t max_cd_steps,
     real_t *restrict buffer_real_t
 );
 void collective_block_cg
@@ -848,6 +951,7 @@ void collective_block_cg
     real_t *restrict precomputedBtB,
     real_t *restrict precomputedCtC, /* should NOT be multiplied by 'w_user' */
     int_t max_cg_steps,
+    real_t *restrict bias_BtX, real_t *restrict bias_X, real_t bias_X_glob,
     real_t *restrict buffer_real_t
 );
 void collective_block_cg_implicit
@@ -876,9 +980,10 @@ void optimizeA_collective_implicit
     size_t U_csr_p[], int_t U_csr_i[], real_t *restrict U_csr,
     real_t *restrict U, int_t cnt_NA_u[],
     bool full_dense_u, bool near_dense_u, bool NA_as_zero_U,
-    real_t lam, real_t w_user,
+    real_t lam, real_t l1_lam, real_t w_user,
     int_t nthreads,
     bool use_cg, int_t max_cg_steps, bool is_first_iter,
+    bool nonneg, int_t max_cd_steps,
     real_t *restrict precomputedBtB, /* will not have lambda with CG */
     real_t *restrict precomputedBeTBe,
     real_t *restrict precomputedBeTBeChol,
@@ -899,8 +1004,10 @@ int_t collective_factors_cold
     real_t *restrict CtCw,
     real_t *restrict col_means,
     int_t k, int_t k_user, int_t k_main,
-    real_t lam, real_t w_main, real_t w_user,
-    bool NA_as_zero_U
+    real_t lam, real_t l1_lam, real_t w_main, real_t w_user,
+    bool scale_lam,
+    bool NA_as_zero_U,
+    bool nonneg
 );
 int_t collective_factors_cold_implicit
 (
@@ -914,8 +1021,10 @@ int_t collective_factors_cold_implicit
     real_t *restrict BeTBeChol,
     real_t *restrict col_means,
     int_t k, int_t k_user, int_t k_item, int_t k_main,
-    real_t lam, real_t w_main, real_t w_user, real_t w_main_multiplier,
-    bool NA_as_zero_U
+    real_t lam, real_t l1_lam,
+    real_t w_main, real_t w_user, real_t w_main_multiplier,
+    bool NA_as_zero_U,
+    bool nonneg
 );
 int_t collective_factors_warm
 (
@@ -933,13 +1042,17 @@ int_t collective_factors_warm
     real_t *restrict Bi, bool add_implicit_features,
     int_t k, int_t k_user, int_t k_item, int_t k_main,
     real_t lam, real_t w_main, real_t w_user, real_t w_implicit,real_t lam_bias,
+    real_t l1_lam, real_t l1_lam_bias,
+    bool scale_lam, bool scale_lam_sideinfo,
     int_t n_max, bool include_all_X,
     real_t *restrict TransBtBinvBt,
+    real_t *restrict BtXbias,
     real_t *restrict BtB,
     real_t *restrict BeTBeChol,
     real_t *restrict BiTBi,
     real_t *restrict CtCw,
     bool NA_as_zero_U, bool NA_as_zero_X,
+    bool nonneg,
     real_t *restrict B_plus_bias
 );
 int_t collective_factors_warm_implicit
@@ -948,11 +1061,12 @@ int_t collective_factors_warm_implicit
     real_t *restrict u_vec, int_t p,
     real_t *restrict u_vec_sp, int_t u_vec_ixB[], size_t nnz_u_vec,
     bool NA_as_zero_U,
+    bool nonneg,
     real_t *restrict col_means,
     real_t *restrict B, int_t n, real_t *restrict C,
     real_t *restrict Xa, int_t ixB[], size_t nnz,
     int_t k, int_t k_user, int_t k_item, int_t k_main,
-    real_t lam, real_t alpha, real_t w_main, real_t w_user,
+    real_t lam, real_t l1_lam, real_t alpha, real_t w_main, real_t w_user,
     real_t w_main_multiplier,
     real_t *restrict BeTBe,
     real_t *restrict BtB,
@@ -1007,6 +1121,7 @@ size_t buffer_size_optimizeA_collective
     bool add_implicit_features, size_t k_main_i,
     size_t nthreads,
     bool use_cg, bool finalize_chol,
+    bool nonneg, bool has_l1,
     bool keep_precomputed,
     bool pass_allocated_BtB,
     bool pass_allocated_CtCw,
@@ -1021,6 +1136,7 @@ size_t buffer_size_optimizeA_collective_implicit
     bool NA_as_zero_U,
     size_t nthreads,
     bool use_cg,
+    bool nonneg, bool has_l1,
     bool pass_allocated_BtB,
     bool pass_allocated_BeTBe,
     bool pass_allocated_BeTBeChol,
@@ -1043,15 +1159,20 @@ void optimizeA_collective
     real_t *restrict U, int_t cnt_NA_u[],
     bool full_dense_u, bool near_dense_u, bool NA_as_zero_U,
     real_t lam, real_t w_user, real_t w_implicit, real_t lam_last,
+    real_t l1_lam, real_t l1_lam_bias,
+    bool scale_lam, bool scale_lam_sideinfo,
     bool do_B,
     int_t nthreads,
     bool use_cg, int_t max_cg_steps, bool is_first_iter,
+    bool nonneg, int_t max_cd_steps,
+    real_t *restrict bias_BtX, real_t *restrict bias_X, real_t bias_X_glob,
     bool keep_precomputed,
     real_t *restrict precomputedBtB,
     real_t *restrict precomputedCtCw,
     real_t *restrict precomputedBeTBeChol,
     real_t *restrict precomputedBiTBi,
     bool *filled_BtB, bool *filled_CtCw, bool *filled_BeTBeChol,
+    bool *CtC_is_scaled,
     real_t *restrict buffer_real_t
 );
 void build_BeTBe
@@ -1106,7 +1227,7 @@ int_t preprocess_sideinfo_matrix
     size_t **U_csc_p, int_t **U_csc_i, real_t *restrict *U_csc,
     int_t *restrict *cnt_NA_u_byrow, int_t *restrict *cnt_NA_u_bycol,
     bool *full_dense_u, bool *near_dense_u_row, bool *near_dense_u_col,
-    bool NA_as_zero_U, int_t nthreads
+    bool NA_as_zero_U, bool nonneg, int_t nthreads
 );
 real_t wrapper_collective_fun_grad
 (
@@ -1140,7 +1261,6 @@ typedef struct data_collective_fun_grad {
     real_t w_main; real_t w_user; real_t w_item;
     int_t nthreads;
     int_t print_every; int_t nfev; int_t niter;
-    bool handle_interrupt;
 } data_collective_fun_grad;
 int_t fit_collective_explicit_lbfgs_internal
 (
@@ -1151,7 +1271,7 @@ int_t fit_collective_explicit_lbfgs_internal
     int_t ixA[], int_t ixB[], real_t *restrict X, size_t nnz,
     real_t *restrict Xfull,
     real_t *restrict weight,
-    bool user_bias, bool item_bias,
+    bool user_bias, bool item_bias, bool center,
     real_t lam, real_t *restrict lam_unique,
     real_t *restrict U, int_t m_u, int_t p,
     real_t *restrict II, int_t n_i, int_t q,
@@ -1180,7 +1300,7 @@ int_t fit_collective_explicit_lbfgs
     int_t ixA[], int_t ixB[], real_t *restrict X, size_t nnz,
     real_t *restrict Xfull,
     real_t *restrict weight,
-    bool user_bias, bool item_bias,
+    bool user_bias, bool item_bias, bool center,
     real_t lam, real_t *restrict lam_unique,
     real_t *restrict U, int_t m_u, int_t p,
     real_t *restrict II, int_t n_i, int_t q,
@@ -1217,8 +1337,10 @@ int_t fit_collective_explicit_als
     int_t ixA[], int_t ixB[], real_t *restrict X, size_t nnz,
     real_t *restrict Xfull,
     real_t *restrict weight,
-    bool user_bias, bool item_bias,
+    bool user_bias, bool item_bias, bool center,
     real_t lam, real_t *restrict lam_unique,
+    real_t l1_lam, real_t *restrict l1_lam_unique,
+    bool scale_lam, bool scale_lam_sideinfo,
     real_t *restrict U, int_t m_u, int_t p,
     real_t *restrict II, int_t n_i, int_t q,
     int_t U_row[], int_t U_col[], real_t *restrict U_sp, size_t nnz_U,
@@ -1228,15 +1350,17 @@ int_t fit_collective_explicit_als
     real_t w_main, real_t w_user, real_t w_item, real_t w_implicit,
     int_t niter, int_t nthreads, bool verbose, bool handle_interrupt,
     bool use_cg, int_t max_cg_steps, bool finalize_chol,
+    bool nonneg, int_t max_cd_steps, bool nonneg_C, bool nonneg_D,
     bool precompute_for_predictions,
     bool include_all_X,
     real_t *restrict B_plus_bias,
     real_t *restrict precomputedBtB,
     real_t *restrict precomputedTransBtBinvBt,
+    real_t *restrict precomputedBtXbias,
     real_t *restrict precomputedBeTBeChol,
+    real_t *restrict precomputedBiTBi,
     real_t *restrict precomputedTransCtCinvCt,
-    real_t *restrict precomputedCtCw,
-    real_t *restrict precomputedBiTBi
+    real_t *restrict precomputedCtCw
 );
 int_t fit_collective_implicit_als
 (
@@ -1247,6 +1371,7 @@ int_t fit_collective_implicit_als
     int_t m, int_t n, int_t k,
     int_t ixA[], int_t ixB[], real_t *restrict X, size_t nnz,
     real_t lam, real_t *restrict lam_unique,
+    real_t l1_lam, real_t *restrict l1_lam_unique,
     real_t *restrict U, int_t m_u, int_t p,
     real_t *restrict II, int_t n_i, int_t q,
     int_t U_row[], int_t U_col[], real_t *restrict U_sp, size_t nnz_U,
@@ -1258,6 +1383,7 @@ int_t fit_collective_implicit_als
     real_t alpha, bool adjust_weight, bool apply_log_transf,
     int_t niter, int_t nthreads, bool verbose, bool handle_interrupt,
     bool use_cg, int_t max_cg_steps, bool finalize_chol,
+    bool nonneg, int_t max_cd_steps, bool nonneg_C, bool nonneg_D,
     bool precompute_for_predictions,
     real_t *restrict precomputedBtB,
     real_t *restrict precomputedBeTBe,
@@ -1268,13 +1394,17 @@ int_t precompute_collective_explicit
     real_t *restrict B, int_t n, int_t n_max, bool include_all_X,
     real_t *restrict C, int_t p,
     real_t *restrict Bi, bool add_implicit_features,
+    real_t *restrict biasB, real_t glob_mean, bool NA_as_zero_X,
     int_t k, int_t k_user, int_t k_item, int_t k_main,
     bool user_bias,
+    bool nonneg,
     real_t lam, real_t *restrict lam_unique,
+    bool scale_lam, bool scale_lam_sideinfo,
     real_t w_main, real_t w_user, real_t w_implicit,
     real_t *restrict B_plus_bias,
     real_t *restrict BtB,
     real_t *restrict TransBtBinvBt,
+    real_t *restrict BtXbias,
     real_t *restrict BeTBeChol,
     real_t *restrict BiTBi,
     real_t *restrict TransCtCinvCt,
@@ -1286,6 +1416,7 @@ int_t precompute_collective_implicit
     real_t *restrict C, int_t p,
     int_t k, int_t k_user, int_t k_item, int_t k_main,
     real_t lam, real_t w_main, real_t w_user, real_t w_main_multiplier,
+    bool nonneg,
     bool extra_precision,
     real_t *restrict BtB,
     real_t *restrict BeTBe,
@@ -1298,6 +1429,7 @@ int_t factors_collective_explicit_single
     real_t *restrict u_vec_sp, int_t u_vec_ixB[], size_t nnz_u_vec,
     real_t *restrict u_bin_vec, int_t pbin,
     bool NA_as_zero_U, bool NA_as_zero_X,
+    bool nonneg,
     real_t *restrict C, real_t *restrict Cb,
     real_t glob_mean, real_t *restrict biasB,
     real_t *restrict U_colmeans,
@@ -1308,10 +1440,13 @@ int_t factors_collective_explicit_single
     real_t *restrict Bi, bool add_implicit_features,
     int_t k, int_t k_user, int_t k_item, int_t k_main,
     real_t lam, real_t *restrict lam_unique,
+    real_t l1_lam, real_t *restrict l1_lam_unique,
+    bool scale_lam, bool scale_lam_sideinfo,
     real_t w_main, real_t w_user, real_t w_implicit,
     int_t n_max, bool include_all_X,
-    real_t *restrict TransBtBinvBt,
     real_t *restrict BtB,
+    real_t *restrict TransBtBinvBt,
+    real_t *restrict BtXbias,
     real_t *restrict BeTBeChol,
     real_t *restrict BiTBi,
     real_t *restrict CtCw,
@@ -1324,11 +1459,12 @@ int_t factors_collective_implicit_single
     real_t *restrict u_vec, int_t p,
     real_t *restrict u_vec_sp, int_t u_vec_ixB[], size_t nnz_u_vec,
     bool NA_as_zero_U,
-    real_t *restrict col_means,
+    bool nonneg,
+    real_t *restrict U_colmeans,
     real_t *restrict B, int_t n, real_t *restrict C,
     real_t *restrict Xa, int_t ixB[], size_t nnz,
     int_t k, int_t k_user, int_t k_item, int_t k_main,
-    real_t lam, real_t alpha, real_t w_main, real_t w_user,
+    real_t lam, real_t l1_lam, real_t alpha, real_t w_main, real_t w_user,
     real_t w_main_multiplier,
     bool apply_log_transf,
     real_t *restrict BeTBe,
@@ -1340,6 +1476,7 @@ int_t factors_collective_explicit_multiple
     real_t *restrict A, real_t *restrict biasA, int_t m,
     real_t *restrict U, int_t m_u, int_t p,
     bool NA_as_zero_U, bool NA_as_zero_X,
+    bool nonneg,
     int_t U_row[], int_t U_col[], real_t *restrict U_sp, size_t nnz_U,
     size_t U_csr_p[], int_t U_csr_i[], real_t *restrict U_csr,
     real_t *restrict Ub, int_t m_ubin, int_t pbin,
@@ -1354,10 +1491,13 @@ int_t factors_collective_explicit_multiple
     real_t *restrict Bi, bool add_implicit_features,
     int_t k, int_t k_user, int_t k_item, int_t k_main,
     real_t lam, real_t *restrict lam_unique,
+    real_t l1_lam, real_t *restrict l1_lam_unique,
+    bool scale_lam, bool scale_lam_sideinfo,
     real_t w_main, real_t w_user, real_t w_implicit,
     int_t n_max, bool include_all_X,
-    real_t *restrict TransBtBinvBt,
     real_t *restrict BtB,
+    real_t *restrict TransBtBinvBt,
+    real_t *restrict BtXbias,
     real_t *restrict BeTBeChol,
     real_t *restrict BiTBi,
     real_t *restrict TransCtCinvCt,
@@ -1370,15 +1510,16 @@ int_t factors_collective_implicit_multiple
     real_t *restrict A, int_t m,
     real_t *restrict U, int_t m_u, int_t p,
     bool NA_as_zero_U,
+    bool nonneg,
     int_t U_row[], int_t U_col[], real_t *restrict U_sp, size_t nnz_U,
     size_t U_csr_p[], int_t U_csr_i[], real_t *restrict U_csr,
     real_t *restrict X, int_t ixA[], int_t ixB[], size_t nnz,
     size_t *restrict Xcsr_p, int_t *restrict Xcsr_i, real_t *restrict Xcsr,
     real_t *restrict B, int_t n,
     real_t *restrict C,
-    real_t *restrict col_means,
+    real_t *restrict U_colmeans,
     int_t k, int_t k_user, int_t k_item, int_t k_main,
-    real_t lam, real_t alpha, real_t w_main, real_t w_user,
+    real_t lam, real_t l1_lam, real_t alpha, real_t w_main, real_t w_user,
     real_t w_main_multiplier,
     bool apply_log_transf,
     real_t *restrict BeTBe,
@@ -1391,6 +1532,7 @@ int_t impute_X_collective_explicit
     int_t m, bool user_bias,
     real_t *restrict U, int_t m_u, int_t p,
     bool NA_as_zero_U,
+    bool nonneg,
     int_t U_row[], int_t U_col[], real_t *restrict U_sp, size_t nnz_U,
     size_t U_csr_p[], int_t U_csr_i[], real_t *restrict U_csr,
     real_t *restrict Ub, int_t m_ubin, int_t pbin,
@@ -1403,10 +1545,12 @@ int_t impute_X_collective_explicit
     real_t *restrict Bi, bool add_implicit_features,
     int_t k, int_t k_user, int_t k_item, int_t k_main,
     real_t lam, real_t *restrict lam_unique,
+    real_t l1_lam, real_t *restrict l1_lam_unique,
+    bool scale_lam, bool scale_lam_sideinfo,
     real_t w_main, real_t w_user, real_t w_implicit,
     int_t n_max, bool include_all_X,
-    real_t *restrict TransBtBinvBt,
     real_t *restrict BtB,
+    real_t *restrict TransBtBinvBt,
     real_t *restrict BeTBeChol,
     real_t *restrict BiTBi,
     real_t *restrict TransCtCinvCt,
@@ -1446,6 +1590,7 @@ int_t topN_new_collective_explicit
     real_t *restrict u_vec_sp, int_t u_vec_ixB[], size_t nnz_u_vec,
     real_t *restrict u_bin_vec, int_t pbin,
     bool NA_as_zero_U, bool NA_as_zero_X,
+    bool nonneg,
     real_t *restrict C, real_t *restrict Cb,
     real_t glob_mean, real_t *restrict biasB,
     real_t *restrict U_colmeans,
@@ -1456,10 +1601,13 @@ int_t topN_new_collective_explicit
     real_t *restrict Bi, bool add_implicit_features,
     int_t k, int_t k_user, int_t k_item, int_t k_main,
     real_t lam, real_t *restrict lam_unique,
+    real_t l1_lam, real_t *restrict l1_lam_unique,
+    bool scale_lam, bool scale_lam_sideinfo,
     real_t w_main, real_t w_user, real_t w_implicit,
     int_t n_max, bool include_all_X,
-    real_t *restrict TransBtBinvBt,
     real_t *restrict BtB,
+    real_t *restrict TransBtBinvBt,
+    real_t *restrict BtXbias,
     real_t *restrict BeTBeChol,
     real_t *restrict BiTBi,
     real_t *restrict CtCw,
@@ -1478,11 +1626,12 @@ int_t topN_new_collective_implicit
     real_t *restrict u_vec, int_t p,
     real_t *restrict u_vec_sp, int_t u_vec_ixB[], size_t nnz_u_vec,
     bool NA_as_zero_U,
-    real_t *restrict col_means,
+    bool nonneg,
+    real_t *restrict U_colmeans,
     real_t *restrict B, real_t *restrict C,
     real_t *restrict Xa, int_t ixB[], size_t nnz,
     int_t k, int_t k_user, int_t k_item, int_t k_main,
-    real_t lam, real_t alpha, real_t w_main, real_t w_user,
+    real_t lam, real_t l1_lam, real_t alpha, real_t w_main, real_t w_user,
     real_t w_main_multiplier,
     bool apply_log_transf,
     real_t *restrict BeTBe,
@@ -1523,6 +1672,7 @@ int_t predict_X_new_collective_explicit
     bool user_bias,
     real_t *restrict U, int_t m_u, int_t p,
     bool NA_as_zero_U, bool NA_as_zero_X,
+    bool nonneg,
     int_t U_row[], int_t U_col[], real_t *restrict U_sp, size_t nnz_U,
     size_t U_csr_p[], int_t U_csr_i[], real_t *restrict U_csr,
     real_t *restrict Ub, int_t m_ubin, int_t pbin,
@@ -1537,10 +1687,13 @@ int_t predict_X_new_collective_explicit
     real_t *restrict Bi, bool add_implicit_features,
     int_t k, int_t k_user, int_t k_item, int_t k_main,
     real_t lam, real_t *restrict lam_unique,
+    real_t l1_lam, real_t *restrict l1_lam_unique,
+    bool scale_lam, bool scale_lam_sideinfo,
     real_t w_main, real_t w_user, real_t w_implicit,
     int_t n_max, bool include_all_X,
-    real_t *restrict TransBtBinvBt,
     real_t *restrict BtB,
+    real_t *restrict TransBtBinvBt,
+    real_t *restrict BtXbias,
     real_t *restrict BeTBeChol,
     real_t *restrict BiTBi,
     real_t *restrict TransCtCinvCt,
@@ -1556,15 +1709,16 @@ int_t predict_X_new_collective_implicit
     /* inputs for factors */
     real_t *restrict U, int_t m_u, int_t p,
     bool NA_as_zero_U,
+    bool nonneg,
     int_t U_row[], int_t U_col[], real_t *restrict U_sp, size_t nnz_U,
     size_t U_csr_p[], int_t U_csr_i[], real_t *restrict U_csr,
     real_t *restrict X, int_t ixA[], int_t ixB[], size_t nnz,
     size_t *restrict Xcsr_p, int_t *restrict Xcsr_i, real_t *restrict Xcsr,
     real_t *restrict B, int_t n,
     real_t *restrict C,
-    real_t *restrict col_means,
+    real_t *restrict U_colmeans,
     int_t k, int_t k_user, int_t k_item, int_t k_main,
-    real_t lam, real_t alpha, real_t w_main, real_t w_user,
+    real_t lam, real_t l1_lam, real_t alpha, real_t w_main, real_t w_user,
     real_t w_main_multiplier,
     bool apply_log_transf,
     real_t *restrict BeTBe,
@@ -1653,6 +1807,7 @@ int_t precompute_offsets_both
     real_t *restrict C, int_t p,
     real_t *restrict D, int_t q,
     real_t *restrict C_bias, real_t *restrict D_bias,
+    real_t *restrict biasB, real_t glob_mean, bool NA_as_zero_X,
     bool user_bias, bool add_intercepts, bool implicit,
     int_t k, int_t k_main, int_t k_sec,
     real_t lam, real_t *restrict lam_unique,
@@ -1734,7 +1889,6 @@ typedef struct data_offsets_fun_grad {
     real_t *buffer_real_t;
     real_t *buffer_mt;
     int_t print_every; int_t nfev; int_t niter;
-    bool handle_interrupt;
 } data_offsets_fun_grad;
 real_t wrapper_offsets_fun_grad
 (
@@ -1752,7 +1906,7 @@ int_t fit_offsets_explicit_lbfgs_internal
     int_t ixA[], int_t ixB[], real_t *restrict X, size_t nnz,
     real_t *restrict Xfull,
     real_t *restrict weight,
-    bool user_bias, bool item_bias,
+    bool user_bias, bool item_bias, bool center,
     bool add_intercepts,
     real_t lam, real_t *restrict lam_unique,
     real_t *restrict U, int_t p,
@@ -1780,7 +1934,7 @@ int_t fit_offsets_explicit_lbfgs
     int_t ixA[], int_t ixB[], real_t *restrict X, size_t nnz,
     real_t *restrict Xfull,
     real_t *restrict weight,
-    bool user_bias, bool item_bias,
+    bool user_bias, bool item_bias, bool center,
     bool add_intercepts,
     real_t lam, real_t *restrict lam_unique,
     real_t *restrict U, int_t p,
@@ -1811,7 +1965,7 @@ int_t fit_offsets_als
     int_t ixA[], int_t ixB[], real_t *restrict X, size_t nnz,
     real_t *restrict Xfull,
     real_t *restrict weight,
-    bool user_bias, bool item_bias, bool add_intercepts,
+    bool user_bias, bool item_bias, bool center, bool add_intercepts,
     real_t lam,
     real_t *restrict U, int_t p,
     real_t *restrict II, int_t q,
@@ -1839,7 +1993,7 @@ int_t fit_offsets_explicit_als
     int_t ixA[], int_t ixB[], real_t *restrict X, size_t nnz,
     real_t *restrict Xfull,
     real_t *restrict weight,
-    bool user_bias, bool item_bias, bool add_intercepts,
+    bool user_bias, bool item_bias, bool center, bool add_intercepts,
     real_t lam,
     real_t *restrict U, int_t p,
     real_t *restrict II, int_t q,
