@@ -27,6 +27,11 @@
             "On the difficulty of evaluating baselines:
             A study on recommender systems."
             arXiv preprint arXiv:1905.01395 (2019).
+        (f) Franc, Vojtech, Vaclav Hlavac, and Mirko Navara.
+            "Sequential coordinate-wise algorithm for the
+            non-negative least squares problem."
+            International Conference on Computer Analysis of Images
+            and Patterns. Springer, Berlin, Heidelberg, 2005.
 
     For information about the models offered here and how they are fit to
     the data, see the files 'collective.c' and 'offsets.c'.
@@ -40,7 +45,7 @@
 
     MIT License:
 
-    Copyright (c) 2020 David Cortes
+    Copyright (c) 2021 David Cortes
 
     All rights reserved.
 
@@ -619,6 +624,8 @@ void tgemm_sp_dense
     int_t nthreads
 )
 {
+    if (m <= 0 || indptr[0] == indptr[m])
+        return;
     real_t *ptr_col;
     #if defined(_OPENMP) && \
                 ( (_OPENMP < 200801)  /* OpenMP < 3.0 */ \
@@ -778,12 +785,12 @@ int_t coo_to_csr_plus_alloc
     *csr_p = (size_t*)malloc(((size_t)m+(size_t)1)*sizeof(size_t));
     *csr_i = (int_t*)malloc(nnz*sizeof(int_t));
     *csr_v = (real_t*)malloc(nnz*sizeof(real_t));
-    if (csr_p == NULL || csr_i == NULL || csr_v == NULL)
+    if (*csr_p == NULL || *csr_i == NULL || *csr_v == NULL)
         return 1;
 
     if (W != NULL) {
         *csr_w = (real_t*)malloc(nnz*sizeof(real_t));
-        if (csr_w == NULL) return 1;
+        if (*csr_w == NULL) return 1;
     }
 
     coo_to_csr(
@@ -909,9 +916,9 @@ void coo_to_csr_and_csc
                     }
                 else
                     for (size_t ix = 0; ix < nnz; ix++) {
-                        csr_w[csr_p[Xcol[ix]] + cnt_byrow[Xcol[ix]]] = W[ix];
-                        csr_v[csr_p[Xcol[ix]] + cnt_byrow[Xcol[ix]]] = Xval[ix];
-                        csr_i[csr_p[Xcol[ix]] + cnt_byrow[Xcol[ix]]++] = Xrow[ix];
+                        csr_w[csr_p[Xrow[ix]] + cnt_byrow[Xrow[ix]]] = W[ix];
+                        csr_v[csr_p[Xrow[ix]] + cnt_byrow[Xrow[ix]]] = Xval[ix];
+                        csr_i[csr_p[Xrow[ix]] + cnt_byrow[Xrow[ix]]++] = Xcol[ix];
                     }
 
             }
@@ -996,11 +1003,10 @@ void row_means_csr(size_t indptr[], real_t *restrict values,
 bool should_stop_procedure = false;
 void set_interrup_global_variable(int_t s)
 {
-    fprintf(stderr, "Error: procedure was interrupted\n");
-    #if !defined(_FOR_R)
-    fflush(stderr);
-    #endif
-    should_stop_procedure = true;
+    #pragma omp critical
+    {
+        should_stop_procedure = true;
+    }
 }
 
 int_t lbfgs_printer_collective
@@ -1026,12 +1032,8 @@ int_t lbfgs_printer_collective
         fflush(stdout);
         #endif
     }
-    if (((data_collective_fun_grad*)instance)->handle_interrupt)
-        signal(SIGINT, set_interrup_global_variable);
-    if (should_stop_procedure) {
-        should_stop_procedure = false;
+    if (should_stop_procedure)
         return 1;
-    }
     return 0;
 }
 
@@ -1058,12 +1060,8 @@ int_t lbfgs_printer_offsets
         fflush(stdout);
         #endif
     }
-    if (((data_offsets_fun_grad*)instance)->handle_interrupt)
-        signal(SIGINT, set_interrup_global_variable);
-    if (should_stop_procedure) {
-        should_stop_procedure = false;
+    if (should_stop_procedure)
         return 1;
-    }
     return 0;
 }
 
@@ -1143,14 +1141,28 @@ void fill_lower_triangle(real_t A[], size_t n, size_t lda)
             A[col + row*lda] = A[row + col*lda];
 }
 
-void print_oom_message(void)
+void print_err_msg(const char *msg)
 {
-    fprintf(stderr, "Error: could not allocate enough memory.\n");
+    fprintf(stderr, msg);
     #ifndef _FOR_R
     fflush(stderr);
     #endif
 }
 
+void print_oom_message(void)
+{
+    print_err_msg("Error: could not allocate enough memory.\n");
+}
+
+void act_on_interrupt(int retval, bool handle_interrupt)
+{
+    if (retval == 3)
+        print_err_msg(" Error: procedure was interrupted.\n");
+    #if !defined(_WIN32) && !defined(_WIN64) && !defined(_MSC_VER)
+    if (!handle_interrupt && retval == 3) kill(getpid(), SIGINT);
+    #endif
+}
+
 #ifdef _FOR_R
 void R_nan_to_C_nan(real_t arr[], size_t n)
 {
@@ -1158,3 +1170,39 @@ void R_nan_to_C_nan(real_t arr[], size_t n)
         arr[ix] = ISNAN(arr[ix])? NAN : arr[ix];
 }
 #endif
+
+long double compensated_sum(real_t *arr, size_t n)
+{
+    long double err = 0.;
+    long double diff = 0.;
+    long double temp;
+    long double res = 0;
+
+    for (size_t ix = 0; ix < n; ix++)
+    {
+        diff = arr[ix] - err;
+        temp = res + diff;
+        err = (temp - res) - diff;
+        res = temp;
+    }
+
+    return res;
+}
+
+long double compensated_sum_product(real_t *restrict arr1, real_t *restrict arr2, size_t n)
+{
+    long double err = 0.;
+    long double diff = 0.;
+    long double temp;
+    long double res = 0;
+
+    for (size_t ix = 0; ix < n; ix++)
+    {
+        diff = arr1[ix]*arr2[ix] - err;
+        temp = res + diff;
+        err = (temp - res) - diff;
+        res = temp;
+    }
+
+    return res;
+}
